
import logging

from django.utils.html import escape
from django.db import models
from django.utils.translation import gettext, pgettext, gettext_lazy as _

from django.views.decorators.csrf import csrf_exempt
from django.core.exceptions import ValidationError
from django.urls import reverse
from django.http import HttpResponse


from rdflib import URIRef
from jsonfield import JSONField

from spkcspider.apps.spider.contents import (
    BaseContent, add_content, VariantType, ActionUrl
)
from spkcspider.apps.spider.helpers import get_settings_func

logger = logging.getLogger(__name__)

try:
    from lru import LRU
    CACHE_FORMS = LRU(256)
except Exception:
    logger.warning("LRU dict failed, use dict instead", exc_info=True)
    CACHE_FORMS = {}


# Create your models here.


class TagLayout(models.Model):
    name = models.SlugField(max_length=255, null=False)
    # shall it be unique for a component?
    unique = models.BooleanField(default=False, blank=True)
    layout = JSONField(
        default=list,
        help_text=_("Field list in JSON format")
    )
    default_verifiers = JSONField(default=list, blank=True)
    usertag = models.OneToOneField(
        "spider_tags.UserTagLayout", on_delete=models.CASCADE,
        related_name="layout", null=True, blank=True
    )
    description = models.TextField(default="", blank=True)

    class Meta(object):
        unique_together = [
            ("name", "usertag")
        ]

    @classmethod
    def localize_name(cls, name):
        _ = pgettext
        return _("content name", "Tag Layout")

    def get_description(self):
        if self.usertag:
            return self.usertag.associated.description
        return self.description

    def full_clean(self, **kwargs):
        # checked with clean
        kwargs.setdefault("exclude", []).append("usertag")
        return super().full_clean(**kwargs)

    def clean(self):
        if self.usertag:
            if TagLayout.objects.filter(usertag=None, name=self.name).exists():
                raise ValidationError(
                    _("Layout exists already"),
                    code="unique"
                )
            self.usertag.full_clean(exclude=["layout"])

    def get_form(self):
        compkey = (self.name, self.usertag.pk if self.usertag else None)
        form = CACHE_FORMS.get(compkey)
        if not form:
            from .forms import generate_form
            form = generate_form("LayoutForm", self.layout)
            CACHE_FORMS[compkey] = form
        return form

    def __repr__(self):
        if self.usertag:
            return "<TagLayout: %s:%s>" % (
                self.name, self.usertag.associated.usercomponent.username
            )
        return "<TagLayout: %s>" % self.name

    def __str__(self):
        if self.usertag:
            return "TagLayout: %s:%s" % (
                self.name, self.usertag.associated.usercomponent.username
            )
        return "TagLayout: %s" % self.name

    def save(self, *args, **kwargs):
        if self.pk:
            # invalidate forms
            old = TagLayout.objects.get(pk=self.pk)
            _id = self.usertag.pk if self.usertag else None
            try:
                del CACHE_FORMS[old.name, _id]
            except KeyError:
                pass
        return super().save(*args, **kwargs)


@add_content
class UserTagLayout(BaseContent):
    # 10 is required for preventing info leak gadgets via component auth
    appearances = [
        {
            "name": "TagLayout",
            "ctype": VariantType.unique.value,
            "strength": 10
        }
    ]
    expose_name = False
    expose_description = True

    @classmethod
    def localize_name(cls, name):
        _ = pgettext
        return _("content name", "Tag Layout")

    def get_content_name(self):
        return "%s: %s" % (
            self.associated.usercomponent.username,
            self.layout.name,
        )

    def localized_description(self):
        """ localize and perform other transforms before rendering to user """
        return gettext(self.associated.description)

    def get_size(self):
        s = super().get_size()
        s += len(str(self.layout.default_verifiers))
        s += len(str(self.layout.layout))
        return s

    def get_template_name(self, scope):
        if scope == "view":
            return 'spider_tags/edit_preview_form.html'
        return super().get_template_name(scope)

    def get_strength_link(self):
        # never allow links to this, elsewise with links is an information
        # disclosure possible
        return 11

    def get_info(self):
        return "%slayout=%s\x1e" % (
            super().get_info(),
            self.layout.name
        )

    def get_form(self, scope):
        if scope in {"add", "update"}:
            from .forms import TagLayoutForm
            return TagLayoutForm
        else:
            return self.layout.get_form()

    def access_view(self, **kwargs):
        _ = gettext
        kwargs.setdefault(
            "legend",
            escape(_("Check \"%s\"") % self.__str__())
        )
        # not visible by default
        kwargs.setdefault("confirm", _("Check"))
        # prevent second button
        kwargs.setdefault("inner_form", False)
        return super().access_view(**kwargs)

    def access_add(self, **kwargs):
        if not hasattr(self, "layout"):
            self.layout = TagLayout(usertag=self)
        return super().access_add(**kwargs)

    def get_form_kwargs(self, **kwargs):
        kwargs["instance"] = self.layout
        return super().get_form_kwargs(**kwargs)

    def access(self, context):
        if context["scope"] == "view":
            context["extra_outer_forms"] = ["request_verification_form"]
        return super().access(context)


@add_content
class SpiderTag(BaseContent):
    _cached_references = None
    tmp_primary_anchor = None
    appearances = [
        {
            "name": "SpiderTag",
            "strength": 0,
            "ctype": VariantType.domain_mode
        },
        {
            "name": "PushedTag",
            "strength": 2,
            "ctype": VariantType.component_feature + VariantType.domain_mode
        }
    ]
    layout = models.ForeignKey(
        TagLayout, related_name="tags", on_delete=models.PROTECT,

    )
    tagdata = JSONField(default=dict, blank=True)
    verified_by = JSONField(default=list, blank=True)
    updateable_by = models.ManyToManyField(
        "spider_base.ReferrerObject", related_name="tags", blank=True
    )
    primary = models.BooleanField(default=False, blank=True)

    expose_name = False
    expose_description = False

    def __str__(self):
        if not self.id:
            return self.localize_name(self.associated.ctype.name)
        if not self.layout.usertag:
            return "%s: <%s>: %s" % (
                self.localize_name("SpiderTag"),
                self.layout.name,
                self.associated.id
            )
        return "%s: <%s: %s>: %s" % (
            self.localize_name("SpiderTag"),
            self.layout.usertag.associated.usercomponent.username,
            self.layout.name,
            self.associated.id
        )

    @classmethod
    def localize_name(cls, name):
        _ = pgettext
        if name == "PushedTag":
            return _("content name", "Allow receiving tags")
        else:
            return _("content name", "Tag")

    @classmethod
    def feature_urls(cls, name):
        return [
            ActionUrl("pushtag", reverse("spider_tags:create-pushtag"))
        ]

    def get_content_description(self):
        return self.layout.get_description()

    def get_content_name(self):
        if not self.layout.usertag:
            return self.layout.name
        return self.layout.usertag.get_content_name()

    def get_template_name(self, scope):
        if scope == "update":
            return 'spider_tags/edit_form.html'
        if scope == "push_update":
            return 'spider_base/edit_form.html'
        return super().get_template_name(scope)

    def get_size(self):
        s = super().get_size()
        s += len(str(self.verified_by))
        s += len(str(self.tagdata))
        return s

    def get_strength_link(self):
        return 0

    def get_abilities(self, context):
        _abilities = set()
        if (
            context["request"].auth_token and
            context["request"].auth_token.referrer
        ):
            if get_settings_func(
                "SPIDER_TAG_VERIFIER_VALIDATOR",
                "spkcspider.apps.spider.functions.clean_verifier"
            )(self, context["request"]):
                _abilities.add("verify")
            if self.updateable_by.filter(
                id=context["request"].auth_token.referrer.id
            ).exists():
                _abilities.add("push_update")

        return _abilities

    def map_data(self, name, field, data, graph, context):
        if (
            field.__class__.__name__ == "AnchorField" and
            field.use_default_anchor
        ):
            if data is None:
                return URIRef(self.get_primary_anchor(graph, context))
        return super().map_data(name, field, data, graph, context)

    @csrf_exempt
    def access_verify(self, **kwargs):
        # full url to result
        verified = kwargs["request"].POST.get("url", "")
        if "://" not in verified:
            return HttpResponse("invalid url", status=400)
        if verified in self.verified_by:
            return self.access_view(**kwargs)
        self.verified_by.append(verified)
        self.clean()
        self.save()

        return self.access_view(**kwargs)

    def access_push_update(self, **kwargs):
        kwargs["legend"] = escape(
            _("Update \"%s\" (push)") % self.__str__()
        )
        kwargs["inner_form"] = False
        return self.access_update(**kwargs)

    def access(self, context):
        if context["scope"] not in {"add", "view"}:
            context["extra_outer_forms"] = ["request_verification_form"]
        return super().access(context)

    def get_form(self, scope):
        from .forms import SpiderTagForm
        if scope == "add":
            return SpiderTagForm
        else:
            return self.layout.get_form()

    def get_references(self):
        if not getattr(self, "layout", None):
            return []
        if self._cached_references is not None:
            return self._cached_references
        form = self.layout.get_form()(
            initial=self.tagdata.copy(),
            instance=self,
            uc=self.associated.usercomponent
        )
        _cached_references, needs_update = form.calc_references(True)
        if needs_update:
            self.associated.save(
                update_fields=[
                    "attached_to_primary_anchor"
                ]
            )
        self._cached_references = _cached_references
        return self._cached_references

    def get_form_kwargs(self, instance=None, **kwargs):
        if kwargs["scope"] == "add":
            ret = super().get_form_kwargs(
                instance=instance,
                **kwargs
            )
            ret["user"] = self.associated.usercomponent.user
        else:
            ret = super().get_form_kwargs(
                **kwargs
            )
            ret["initial"] = self.tagdata.copy()
            ret["request"] = kwargs["request"]
            ret["uc"] = self.associated.usercomponent
        return ret

    def encode_verifiers(self):
        return "".join(
            map(
                lambda x: "verified_by={}\x1e".format(x),
                self.verified_by
            )
        )

    def get_info(self):
        return "{}{}tag={}\x1e".format(
            super().get_info(unique=self.primary, unlisted=False),
            self.encode_verifiers(),
            self.layout.name
        )

    def save(self, *args, **kwargs):
        super().save(*args, **kwargs)
        self._cached_references = None




import logging
from django.utils.timezone import now

from django.contrib import admin
from django.contrib.auth import get_permission_codename

from .models import DataVerificationTag
from .validate import verify_tag

logger = logging.getLogger(__name__)


@admin.register(DataVerificationTag)
class DataVerificationTagAdmin(admin.ModelAdmin):
    view_on_site = True
    list_display = ('hash', "checked", 'verification_state')
    search_fields = ('hash', "checked", 'verification_state')
    fields = [
        'created', 'modified', 'checked', 'dvfile', 'verification_state',
        'note'
    ]
    readonly_fields = ['created', 'modified', "checked"]

    def get_form(self, request, obj=None, **kwargs):
        opts = self.opts
        codename = get_permission_codename('change', opts)
        if not request.user.has_perm("%s.%s" % (opts.app_label, codename)):
            self.readonly_fields = [
                'created', 'modified', 'checked', 'dvfile'
            ]
        return super().get_form(request, obj, **kwargs)

    def has_module_permission(self, request):
        return True

    def has_view_permission(self, request, obj=None):
        opts = self.opts
        if request.user.has_perm("%s.%s" % (opts.app_label, 'can_verify')):
            return True
        return super().has_view_permission(request, obj)

    def has_change_permission(self, request, obj=None):
        opts = self.opts
        if request.user.has_perm("%s.%s" % (opts.app_label, 'can_verify')):
            return True
        return super().has_change_permission(request, obj)

    def save_form(self, request, form, change):
        if 'verification_state' in form.changed_data:
            form.instance.checked = now()
        ret = super().save_form(request, form, change)

        verify_tag(
            tag=form.instance, ffrom="admin"
        )
        return ret

    # def save_model(self, request, obj, form, change):
    #    """
    #    Given a model instance save it to the database.
    #    """
    #    if 'verification_state' in form.changed_data:
    #        obj.checked = now()
    #    ret = super().save_model(request, obj, form, change)
    #    obj.callback()
    #    return ret



__all__ = ("PushTagView",)

from django.conf import settings
from django.urls import reverse
from django.http import Http404
from django.shortcuts import get_object_or_404
from django.http import JsonResponse, HttpResponseRedirect

from django.views.decorators.csrf import csrf_exempt
from django.utils.decorators import method_decorator
from django.views.generic.edit import FormView

from spkcspider.apps.spider.views import UCTestMixin
from spkcspider.apps.spider.helpers import get_settings_func
from spkcspider.apps.spider.models import (
    AuthToken
)
from .models import SpiderTag
from .forms import SpiderTagForm


class PushTagView(UCTestMixin, FormView):
    model = SpiderTag
    form_class = SpiderTagForm
    variant = None
    object = None

    @method_decorator(csrf_exempt)
    def dispatch(self, request, *args, **kwargs):
        try:
            return super().dispatch(request, *args, **kwargs)
        except Http404:
            return get_settings_func(
                "SPIDER_RATELIMIT_FUNC",
                "spkcspider.apps.spider.functions.rate_limit_default"
            )(self, request)

    def get_form_kwargs(self):
        kwargs = super().get_form_kwargs()
        kwargs["instance"] = self.object
        kwargs["user"] = self.usercomponent.user
        return kwargs

    def get_usercomponent(self):
        self.request.auth_token = get_object_or_404(
            AuthToken,
            token=self.request.GET.get("token", None),
            referrer__isnull=False
        )
        return self.request.auth_token.usercomponent

    def test_func(self):
        self.variant = self.usercomponent.features.filter(
            name="PushedTag"
        ).first()
        # can only access feature if activated
        return bool(self.variant)

    def options(self, request, *args, **kwargs):
        ret = super().options()
        ret["Access-Control-Allow-Origin"] = "*"
        ret["Access-Control-Allow-Methods"] = "POST, GET, OPTIONS"
        return ret

    def post(self, request, *args, **kwargs):
        # not yet confirmed
        self.object = self.model.static_create(
            token_size=getattr(settings, "TOKEN_SIZE", 30),
            associated_kwargs={
                "usercomponent": self.usercomponent,
                "ctype": self.variant
            }
        )
        # return super().post(request, *args, **kwargs)
        form = self.get_form()
        if form.is_valid():
            return self.form_valid(form)
        else:
            return self.form_invalid(form)

    def render_to_response(self, context):
        ret = JsonResponse(
            {
                "layout": [
                    i.name
                    for i in context["form"].fields["layout"].queryset.all()
                ]
            }
        )
        # allow cors requests for accessing data
        ret["Access-Control-Allow-Origin"] = \
            self.request.auth_token.referrer.host
        return ret

    def form_valid(self, form):
        self.object = form.save(commit=False)
        self.object.clean()
        self.object.save()
        form.save_m2m()
        self.object.updateable_by.add(self.request.auth_token.referrer)
        assert(self.object.associated.id)
        assert(self.object.associated.token)
        ret = HttpResponseRedirect(
            redirect_to="{}?token={}".format(
                reverse(
                    "spider_base:ucontent-access",
                    kwargs={
                        "token": self.object.associated.token,
                        "access": "push_update"
                    }
                ),
                self.request.auth_token.token
            )
        )
        ret["Access-Control-Allow-Origin"] = \
            self.request.auth_token.referrer.host
        return ret



# Generated by Django 2.1.1 on 2018-09-22 22:11

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('spider_accounts', '0002_auto_20180810_1433'),
    ]

    operations = [
        migrations.AddField(
            model_name='spideruser',
            name='quota',
            field=models.PositiveIntegerField(blank=True, default=None, help_text='Quota in Bytes, null to use standard', null=True),
        ),
    ]



from django import template

register = template.Library()


@register.simple_tag(takes_context=True)
def list_parameters(context, name, *args):
    ret = set(context["request"].GET.getlist(name))
    ret.update(context["request"].POST.getlist(name))
    for i in args:
        ret.update(i)
    return ret


@register.simple_tag(takes_context=True)
def get_searchpath(context, page=None, pagename="page"):
    if "searchpath" in context:
        path = context["searchpath"]
    else:
        path = context["request"].path
    search_get = ""
    if "sanitized_GET" in context:
        search_get = context["sanitized_GET"]
    if page:
        search_get = "{}{}={}&".format(
            search_get, pagename, page
        )
    return "{}?{}".format(
        path.rstrip("?&"),
        search_get
    )



# Generated by Django 2.2 on 2019-04-25 22:07

from django.db import migrations, models
import django.db.models.deletion
import jsonfield.fields
import spkcspider.apps.spider_keys.models


class Migration(migrations.Migration):

    dependencies = [
        ('spider_keys', '0001_initial'),
    ]
    replaces = [
        ('spider_keys', '0002_0002_20190426_squashed'),
    ]

    operations = [
        migrations.CreateModel(
            name='AnchorServer',
            fields=[
                ('id', models.BigAutoField(editable=False, primary_key=True, serialize=False)),
                ('new_url', models.URLField(blank=True, help_text='Url to new anchor (in case this one is superseded)', max_length=400, null=True)),
                ('old_urls', jsonfield.fields.JSONField(blank=True, default=list, help_text='Superseded anchor urls')),
            ],
            options={
                'abstract': False,
                'default_permissions': (),
            },
        ),
        migrations.AlterField(
            model_name='publickey',
            name='key',
            field=models.TextField(validators=[spkcspider.apps.spider_keys.models.valid_pkey_properties]),
        ),
        migrations.RemoveField(
            model_name='publickey',
            name='note',
        ),
        migrations.CreateModel(
            name='AnchorKey',
            fields=[
                ('id', models.BigAutoField(editable=False, primary_key=True, serialize=False)),
                ('signature', models.CharField(help_text='Signature of Identifier (hexadecimal-encoded)', max_length=1024)),
                ('key', models.OneToOneField(help_text='"Public Key"-Content for signing identifier. It is recommended to use different keys for signing and encryption. Reason herefor is, that with a change of the signing key the whole anchor gets invalid and the signing key should be really carefully saved away. In contrast the encryption keys can be easily exchanged and should be available for encryption', on_delete=django.db.models.deletion.CASCADE, related_name='anchorkey', to='spider_keys.PublicKey')),
            ],
            options={
                'abstract': False,
                'default_permissions': (),
            },
        ),
    ]



__all__ = (
    "create_b64_token", "create_b64_id_token", "get_settings_func",
    "extract_app_dicts", "add_by_field", "prepare_description",
    "merge_get_url", "add_property", "is_decimal",
    "extract_host", "get_hashob", "aesgcm_scrypt_cryptor",
    "aesgcm_pbkdf2_cryptor", "get_requests_params",
    "literalize", "field_to_python"
)


import os
import re
import base64
import logging
import inspect
from hashlib import pbkdf2_hmac
from urllib.parse import urlsplit, urlunsplit, parse_qs, urlencode, urljoin

from functools import lru_cache
from importlib import import_module

from rdflib import Literal, BNode, XSD, RDF, URIRef

from django.conf import settings
from django.core.exceptions import ValidationError
from django.utils.translation import gettext_lazy as _
from django.utils.encoding import force_bytes
from django.forms import BoundField, Field

from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives.kdf.scrypt import Scrypt
from cryptography.hazmat.primitives.ciphers.aead import AESGCM

from .constants import MAX_TOKEN_SIZE, spkcgraph, host_tld_matcher

# for not spamming sets
_empty_set = frozenset()
# for not spamming dicts
_empty_dict = dict()

_pbkdf2_params = {
    "iterations": 120000,
    "hash_name": "sha512",
    "dklen": 32
}

_Scrypt_params = {
    "length": 32,
    "n": 2**14,
    "r": 16,
    "p": 2
}


def get_hashob():
    return hashes.Hash(
        settings.SPIDER_HASH_ALGORITHM, backend=default_backend()
    )


def aesgcm_scrypt_cryptor(pw, salt=None, params=_Scrypt_params):
    if salt is None:
        salt = settings.SECRET_KEY
    salt = force_bytes(salt)

    return AESGCM(
        Scrypt(
            salt=salt,
            backend=default_backend(),
            **params
        ).derive(pw[:128].encode("utf-8"))
    )


def aesgcm_pbkdf2_cryptor(pw, salt=None, params=_pbkdf2_params):
    if salt is None:
        salt = settings.SECRET_KEY
    salt = force_bytes(salt)

    return AESGCM(
        pbkdf2_hmac(
            password=pw[:128].encode("utf-8"),
            salt=salt,
            **params
        )
    )


def is_decimal(inp, precision=None, allow_sign=False):
    prec_start = None
    if len(inp) == 0:
        return False
    if inp[0] == "-" and allow_sign:
        inp = inp[1:]
    for count, i in enumerate(inp):
        if i == "." and prec_start is None:
            prec_start = count
        elif not i.isdigit():
            return False
    return (
        None in (precision, prec_start) or len(inp)-prec_start-1 <= precision
    )


def field_to_python(value):
    if isinstance(value, BoundField):
        data = value.initial
        if value.form.is_bound:
            value = value.field.bound_data(value.data, data)
        else:
            value = data
    return value


def literalize(
    ob=None, datatype=None, use_uriref=None, domain_base=""
):
    if isinstance(ob, BoundField):
        if not datatype:
            datatype = getattr(ob.field, "spkc_datatype", None)
        if use_uriref is None:
            use_uriref = getattr(ob.field, "spkc_use_uriref", None)
        ob = field_to_python(ob)
    elif isinstance(datatype, BoundField):
        if use_uriref is None:
            use_uriref = getattr(datatype.field, "spkc_use_uriref", None)
        datatype = getattr(datatype.field, "spkc_datatype", None)
    elif isinstance(datatype, Field):
        if use_uriref is None:
            use_uriref = getattr(datatype, "spkc_use_uriref", None)
        datatype = getattr(datatype, "spkc_datatype", None)
    if ob is None:
        return RDF.nil
    if hasattr(ob, "get_absolute_url"):
        if not datatype:
            datatype = spkcgraph["hashableURI"]
        if use_uriref is None:
            use_uriref = True
        ob = ob.get_absolute_url()
    elif isinstance(ob, str) and not datatype:
        datatype = XSD.string
    if use_uriref:
        return URIRef(urljoin(domain_base, ob))
    return Literal(ob, datatype=datatype)


def add_property(
    graph, name, ref=None, ob=None, literal=None, datatype=None,
    iterate=False
):
    value_node = BNode()
    if ref:
        graph.add((
            ref, spkcgraph["properties"],
            value_node
        ))
    graph.add((
        value_node, spkcgraph["name"],
        Literal(name, datatype=XSD.string)
    ))
    if literal is None:
        literal = getattr(ob, name)
    if not iterate:
        literal = [literal]
    for l in literal:
        graph.add((
            value_node, spkcgraph["value"],
            Literal(l, datatype=datatype)
        ))
    if not literal:
        graph.set((value_node, spkcgraph["value"], RDF.nil))
    return value_node


@lru_cache(maxsize=None)
def get_settings_func(*args, default=None, exclude=frozenset()):
    if default is None:
        names, default = args[:-1], args[-1]
    filterpath = None
    for name in names:
        if getattr(settings, name, None) is not None:
            filterpath = getattr(settings, name)
            break
    if filterpath is None:
        filterpath = default
    # note: cannot distinguish between False and 0 and True and 1
    if filterpath in exclude:
        raise ValueError("invalid setting")
    if callable(filterpath):
        return filterpath
    elif isinstance(filterpath, bool):
        return lambda *args, **kwargs: filterpath
    else:
        module, name = filterpath.rsplit(".", 1)
        return getattr(import_module(module), name)


def add_by_field(dic, field="__name__"):
    def _func(klass):
        # should be able to block real object
        if "{}.{}".format(klass.__module__, klass.__qualname__) not in getattr(
            settings, "SPIDER_BLACKLISTED_MODULES", _empty_set
        ):
            dic[getattr(klass, field)] = klass

        return klass
    return _func


def extract_app_dicts(app, name, fieldname=None):
    ndic = {}
    for (key, value) in getattr(app, name, _empty_dict).items():
        if inspect.isclass(value):
            if "{}.{}".format(
                value.__module__, value.__qualname__
            ) not in getattr(
                settings, "SPIDER_BLACKLISTED_MODULES", _empty_set
            ):
                if fieldname:
                    setattr(value, fieldname, key)
                ndic[key] = value
        else:
            # set None if Module is not always available
            if value and value not in getattr(
                settings, "SPIDER_BLACKLISTED_MODULES", _empty_set
            ):
                module, name = value.rsplit(".", 1)
                value = getattr(import_module(module), name)
                if fieldname:
                    setattr(value, fieldname, key)
                ndic[key] = value
    return ndic


def create_b64_token(size=None):
    if not size:
        from django.conf import settings
        size = getattr(settings, "TOKEN_SIZE", 30)
    if size > MAX_TOKEN_SIZE:
        logging.warning("Nonce too big")
    return base64.urlsafe_b64encode(
        os.urandom(size)
    ).decode('ascii').rstrip("=")


def create_b64_id_token(id, sep="_", size=None):
    return sep.join((hex(id)[2:], create_b64_token(size)))


_cleanstr = re.compile(r'<+.*>+')
_whitespsplit = re.compile(r'\s+')


def prepare_description(raw_html, amount=0):
    text = _cleanstr.sub(' ', raw_html).\
        replace("<", " ").replace(">", " ").strip()
    return _whitespsplit.split(text, amount)


_check_scheme = re.compile(r'^[a-z]+://', re.I)


def extract_host(url):
    url = url.lstrip(":/")
    if _check_scheme.search(url) is None:
        urlparsed = urlsplit("://".join(("https", url)))
    else:
        urlparsed = urlsplit(url)
    return "://".join(urlparsed[:2])


def merge_get_url(_url, **kwargs):
    if not _url.isprintable():
        raise ValidationError(
            _("Url contains control characters"),
            code="control_characters"
        )
    _url = _url.lstrip(":/")
    if _check_scheme.search(_url) is None:
        urlparsed = urlsplit("://".join(("https", _url)))
    else:
        urlparsed = urlsplit(_url)
    _strip = []
    for i in kwargs.keys():
        if not kwargs[i]:
            _strip.append(i)
    GET = parse_qs(urlparsed.query)
    GET.update(kwargs)
    for item in _strip:
        GET.pop(item, None)
    ret = urlunsplit((*urlparsed[:3], urlencode(GET, doseq=True), ""))
    return ret


def get_requests_params(url):
    _url = host_tld_matcher.match(url)
    if not _url:
        raise ValidationError(
            _("Invalid URL: \"%(url)s\""),
            code="invalid_url",
            params={"url": url}
        )
    _url = _url.groupdict()
    mapper = settings.SPIDER_REQUEST_KWARGS_MAP
    return (
        mapper.get(
            _url["host"],
            mapper.get(
                _url["tld"],  # maybe None but then fall to retrieval 3
                mapper[b"default"]
            )
        ),
        get_settings_func(
            "SPIDER_INLINE",
            "spkcspider.apps.spider.functions.clean_spider_inline",
            exclude=frozenset({True})
        )(_url["host"])
    )
